%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Matlab code (miscellanea)
% Coded by: Eleftherios Giovanis 
% File: STAR_Matlab_Genetic_Algo.m
%
% Paper: ``Smoothing STAR models with OLS and Genetic algorithms Optimization''
% Munich Personal RePEc Archive (MPRA) Paper No. 24660, 2010. 
%
% Smoothing transition functions with ordinary least squares and genetic 
% algorithms optimization.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Load input data
% load file.mat % load the file with data

nforecast = 20;
y         = data(1:end-nforecast,1);
t         = length(y);
d         = 0;   % CHANGE
x         = lagmatrix(y,d);
iii       = 1;
stdev     = std(y);

% Some initial values for parameters c, gamma and b
% c     = 0;
% gamma = 1;
be      = 2;

% Choose STAR function
model   = 1;

% Set up the population size
popsize = 30;
pc      = 0.2;
pm      = 0.01;
lb_c    = 0.03;  % bounds of the parameters c to be optimized
ub_c    = -0.03;

lb_gamma = 1; % bounds of the parameters gamma to be optimized
ub_gamma = 5;

lb_b = 0.5;   % bounds of the parameters b to be optimized
ub_b = 2;

% Set up random population for c
% Range_c = repmat((ub_c-lb_c),[popsize 1]);
  Lower_c = repmat(lb_c, [popsize 1]);
% pop_c   = rand(popsize,1).*Range_c+Lower_c;

% Set up random population for gamma
% Range_gamma = repmat((ub_gamma-lb_gamma),[popsize 1]);
  Lower_gamma = repmat(lb_gamma,[popsize 1]);
% pop_gamma   = rand(popsize,1).*Range_gamma + Lower_gamma;

% Set up random population for b
Range_b = repmat((ub_b-lb_b),[popsize 1]);
Lower_b = repmat(lb_b, [popsize 1]);
pop_b   = rand(popsize,1).*Range_b + Range_b;

% For ESTAR, GBELL_STAR and GAUSS_STAR is better to initialize as
%%  lb = min(min(x));  % bounds of the parameters c to be optimized
%%  ub = max(max(x));

%%  if model==4
%%     pop = [pop_c pop_gamma pop_b];
%%     else
%%     pop = [pop_gamma pop_c];
%%  end

%%  Range = repmat((ub-lb),[popsize chromlength]);
%%  Lower = repmat(lb, [popsize chromlength]);
%%  pop   = rand(popsize,chromlength).* Range + Lower;

Range_c     = repmat((ub_c-lb_c),[popsize 1]);
pop_c       = rand(popsize,1).* Range_c;
Range_gamma = repmat((ub_gamma-lb_gamma),[popsize 1]);
pop_gamma   = rand(popsize,1).* Range_gamma;

%%%  end initialization  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if model==4
   pop = [pop_c pop_gamma pop_b];
else
   pop = [pop_gamma pop_c];
end;

Chromosome = pop;
for p = 1:iii
    ylag(:,p) = lagmatrix(y,p);
end;
% te = length(ylag);

ylag = ylag(iii+d+1:t,:);
y    = y(iii+d+1:t,:);
x    = x(iii+d+1:t,:);

for iterations =1:10  % Change
    [nkk,~]    = size(pop);
    Chromosome = pop;
    for jj = 1:nkk
       X1 = [ones(size(y)) ylag];
       
       if model==1 % ESTAR (exponential)
           ESTAR_1 = 1-exp(-Chromosome(jj,1)./stdev.*((x-Chromosome(jj,2)).^2));
           Y       = X1;
           for p=1:iii+1
               h(:,p) = ESTAR_1.*Y(:,p);
           end;
        
       elseif model==2 % LSTAR (logistic)
           LSTAR_1 = 1./(1+exp(-Chromosome(jj,1).*(x-Chromosome(jj,2))));
           Y       = X1;
           for p=1:iii+1
               h(:,p) = LSTAR_1.*Y(:,p);
           end;
           
       elseif model==3 % TAHN_STAR (tangent hyperbolic)
           TSTAR_1 = (2./(1+exp(-2*Chromosome(jj,1).*(x-Chromosome(jj,2))))-1);
           Y       = X1;
           for p=1:iii+1
               h(:,p) = TSTAR_1.*Y(:,p);
           end;
     
       elseif model==4 % Generalized Bell function
           GBELL_STAR_1 = (1./(1+abs((x-Chromosome(jj,1))/Chromosome(jj,2)).^2*Chromosome(jj,3)));
           Y            = X1;
           for p=1:iii+1
               h(:,p) = GBELL_STAR_1.*Y(:,p);
           end;
         
       elseif model==5 % Gaussian
           GAUSS_STAR_1 = exp(-(x-Chromosome(jj,1)).^2/(2*Chromosome(jj,2)^2));
           Y            = X1;
           for p=1:iii+1
               h(:,p) = GAUSS_STAR_1.*Y(:,p);
           end;
       end;

       X     = [X1 h];
       X2    = [X1 x];
       bols  = inv(X'*X)*X'*y;
       yy1   = X*bols;
       e     = yy1-y;
       error = sum(sum(e.^2));
       object_value(jj,:) = 1/2*mse(e);
   end;   % end for jj=1:nkk

   fitness_value = object_value;
   total_fit     = sum(fitness_value);
   fitness_value = fitness_value/total_fit;
   fitness_value = cumsum(fitness_value);
   [nkk,nii]     = size(pop);
   ms            = sort(rand(nkk,1));

   fitin = 1;
   newin = 1;
   while newin<=nkk
      if(ms(newin))<fitness_value(fitin)
          newpop(newin,:) = pop(fitin,:);
          newin = newin+1;
      else
        fitin = fitin+1;
      end;
   end;

%  Crossover between chromosomes
   pop = newpop;

   length_chrom = size(pop,2);
   cp = ceil(rand(size(pop,1)/2,1)*(length_chrom-1));
   cp = cp.*(rand(size(cp))<pc);
  
   for i = 1:length(cp);
        newpop([2*i-1 2*i],:) = [pop([2*i-1 2*i],1:cp(i)) ...
        pop([2*i 2*i-1],cp(i)+1:length_chrom)];
   end;

   pop         = newpop;
   mutated_pop = find(rand(size(pop))<pm);
   newpop      = pop;
   newpop(mutated_pop) = 1-pop(mutated_pop);

   % Finding the best individual
   pop = newpop;
   best_individual = pop(1,:);
   best_fit        = fitness_value(1);

   for jj=2:nkk
       if fitness_value(jj)<best_fit
          best_individual = pop(jj,:);
          best_fit        = fitness_value(jj);
       end;
   end;

   disp('iterations')
   disp(iterations)
   array_best(iterations) = best_fit;
   index_iter(iterations) = iterations;

end;  %  end iterations

best_gen = best_individual;
if model==4
   c     = best_gen(:,1);
   gamma = best_gen(:,2);
   be    = best_gen(:,3);
else
   gamma = best_gen(:,1);
   c = best_gen(:,2);
end;

if model==1     % ESTAR (exponential)
     ESTAR_2 = 1-exp(-gamma.*((x-c).^2));
     Y       = X1;
     f       = y(end,:)*ESTAR_2(end,:);
     for p=1:iii+1
         h(:,p) = ESTAR_2.*Y(:,p);
     end;

   elseif model==2 % LSTAR (logistic)
      LSTAR_2 = 1./(1+exp(-gamma.*(x-c)));
      Y = X1;
      f = y(end,:)*LSTAR_2(end,:);
      for p=1:iii+1
          h(:,p) = LSTAR_2.*Y(:,p);
      end;

   elseif model==3 % TAHN_STAR (tangent hyperbolic)
      TSTAR_2 = (2./(1+exp(-2*gamma.*(x-c)))-1);
      Y       = X1;
      f       = y(end,:)*TSTAR_2(end,:);
      for p=1:iii+1
          h(:,p) = TSTAR_2.*Y(:,p);
      end;

   elseif model==4 % Generalized Bell function
      GBELL_STAR_2 = (1./(1+abs((x-c)/gamma).^2*be));
      Y            = X1;
      f            = y(end,:)*GBELL_STAR_2(end,:);
      for p=1:iii+1
          h(:,p) = GBELL_STAR_2.*Y(:,p);
      end;
 
  elseif model==5 % Gaussian
     GAUSS_STAR_2 = exp(-(x-c).^2/(2*gamma^2));
     Y            = X1;
     f            = y(end,:)*GAUSS_STAR_2(end,:);
     for p=1:iii+1
         h(:,p) = GAUSS_STAR_2.*Y(:,p);
     end;
end;   % end if

X_new    = [Y h];
[nk,ni]  = size(X_new);
bols_new = inv(X_new'*X_new)*X_new'*y;
res      = y-X_new*bols_new;
s2       = (y-X_new*bols_new)'*(y-X_new*bols_new)/(nk-ni);
Vb       = s2*inv(X_new'*X_new); % Get the variance-covariance matrix
se       = sqrt(diag(Vb));       % Get coefficient standard errors
tstudent = bols_new./se;
yf       = X_new*bols;

[~,pValue,ARCHstat,~] = archtest(res,2,0.05);
% Heteroskedasticity (ARCH) test
e2 = res.*res;
x2 = x.*x;
v  = e2-x2*(e2'/x2')';
e2 = e2-mean(e2)';
te = length(res(:,1))*(1-(v'*v)/(e2'*e2));
ht = 1-chi2cdf(te,length(x(1,:)));

% Ljung_Box statistic for autocorrelation
[~,p_Jung,Qstat,~] = lbqtest(res,2,0.05);
for kkk=1:nk
    if yf(kkk,:)>0
       S_in_sample(kkk,:) = 1;
    elseif yf(kkk,:)<0
       S_in_sample(kkk,:) = 0;
    end;
end;

Actual_positive_in_sample = find(y==1);
Actual_negative_in_sample = find(y==0);

Predicted_positive_in_sample = find(S_in_sample==1);
Predicted_negative_in_sample = find(S_in_sample==0);

Sum_actual_positive_in_sample = length(Actual_positive_in_sample);
Sum_actual_negative_in_sample = length(Actual_negative_in_sample);

Sum_predicted_positive_in_sample = length(Predicted_positive_in_sample);
Sum_predicted_negative_in_sample = length(Predicted_negative_in_sample);

Total_predicted_in_sample = find(y==S_in_sample);
Perc                      = (length(Total_predicted_in_sample)/nk)*100;

for ii=nforecast:-1:1
    y = data(1:end-ii,1);
    x = lagmatrix(y,d);
    clear ylag
    for p=1:iii
        ylag(:,p) = lagmatrix(y,p);
    end;
    
    te   = length(ylag);
    t    = length(y);
    ylag = ylag(iii+d+1:t,:);
    y    = y(iii+d+1:t,:);
    x    = x(iii+d+1:t,:);
    X1   = [ones(size(y)) ylag];
    
    if model==1 % ESTAR (exponential)
       ESTAR_1 = 1-exp(-gamma.*((x-c).^2));
       Y       = X1;
       f       = y(end,:)*ESTAR_1(end,:);
       clear h
       for p=1:iii+1
           h(:,p) = ESTAR_1.*Y(:,p);
       end;

    elseif model==2 % LSTAR (logistic)
         LSTAR_1 = 1./(1+exp(-gamma.*(x-c)));
         Y       = X1;
         f       = y(end,:)*LSTAR_1(end,:);
         clear h
         for p=1:iii+1
             h(:,p) = LSTAR_1.*Y(:,p);
         end;

    elseif model==3 % TAHN_STAR (tangent hyperbolic)
         TSTAR_1 = (2./(1+exp(-2*gamma.*(x-c)))-1);
         Y       = X1;
         f       = y(end,:)*TSTAR_1(end,:);
         clear h
         for p=1:iii+1
             h(:,p) = TSTAR_1.*Y(:,p);
         end;

    elseif model==4 % Generalized Bell function
         GBELL_STAR_1 = (1./(1+abs((x-c)/gamma).^2*be));
         Y            = X1;
         f            = y(end,:)*GBELL_STAR_1(end,:);
         clear h
         for p=1:iii+1
             h(:,p) = GBELL_STAR_1.*Y(:,p);
         end;

    elseif model==5 % Gaussian
         GAUSS_STAR_1 = exp(-(x-c).^2/(2*gamma^2));
         Y            = X1;
         f            = y(end,:)*GAUSS_STAR_1(end,:);
         clear h
         for p=1:iii+1
             h(:,p) = GAUSS_STAR_1.*Y(:,p);
          end;
    end;
    X          = [Y h];
    yhat(ii,:) = [X(end,1) y(end,:) X(end,3) f]*bols;
end;  %  end for loop ii

for iii=1:nforecast
     yfore(iii,:) = yhat(end-iii+1,:);
end;

test_y  = data(end-nforecast+1:end,1);

X3      = [X1 h];
predict = X3*bols;
tt      = length(test_y);

for kkk=1:tt
    if test_y(kkk,:)>0
       S_out_of_sample(kkk,:) = 1;
    elseif test_y(kkk,:)<0
       S_out_of_sample(kkk,:) = 0;
    end;
end;

for kkk=1:tt
    if yfore(kkk,:)>0
       S(kkk,:) = 1;
    elseif yfore(kkk,:)<0
       S(kkk,:) = 0;
    end;
end;

Actual_positive_out_of_sample = find(S_out_of_sample==1);
Actual_negative_out_of_sample = find(S_out_of_sample==0);

Predicted_positive_out_of_sample = find(S==1);
Predicted_negative_out_of_sample = find(S==0);

Total_predicted_out_of_sample = find(S_out_of_sample==S);
Perc_out_sample               = (length(Total_predicted_out_of_sample)/tt)*100;

figure,
plot(y(1:end-nforecast,:),'-r'); 
hold on; 
plot(yf,'-b');
xlabel('Periods')
ylabel('Values')
h1 = legend('Actual','forecasts');

title('Out of sample forecasts')
figure,
plot(test_y,'-r'); 
hold on; 
plot(yfore,'-b');
figure, 
plot(index_iter,array_best);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The initialization for ESTAR, LSTAR and TSTAR is
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

lb = min(min(x));   % bounds of the parameters c to be optimized
ub = max(max(x));

if model==4
   pop = [pop_c pop_gamma pop_b];
else
   pop = [pop_gamma pop_c];
end;

Range = repmat((ub-lb),[popsize chromlength]);
Lower = repmat(lb,[popsize chromlength]);
pop   = rand(popsize,chromlength).* Range + Lower;